--Object Randomizer
--Made by Nightwolf

local urmain = assert(_URANDOM,"URANDOM: variables.lua isn't loaded properly!")
local vars = urmain.var.objrandom

//Only objects with these flags are affected by type change on limited mode
local limitedflags = MF_ENEMY|MF_SHOOTABLE|MF_MONITOR

//Won't be randomized even in full mode
local flagblacklist = MF_BOXICON|MF_SPRING|MF_AMBIENT|MF_BOSS

//Has to be a standard lua array starting at index 1 and without "holes" (nil variables between normal ones)
//If there are "holes", you have to specify table size in 'size' argument
local shufflearray = urmain.shufflearray

local ObjLoading = false

addHook("NetVars",function(net)
	ObjLoading = net(ObjLoading)
end)

local function isMoPickup(mo)
	return (mo.type == MT_RING or mo.type == MT_COIN or (mo.type <= MT_BLUESPHERE and mo.type >= MT_EMERALD7) or mo.flags & MF_MONITOR)
end

//Get objects for position swapping
local function GetObjectsSwap()
	local motab = {}
	local postab = {}
	for mo in mobjs.iterate() do
		if mo.type == MT_PLAYER or mo.type == MT_SKYBOX then continue end
		if mo.flags & flagblacklist != 0 then continue end
		if mo.type >= MT_TELEPORTMAN and mo.type <= MT_SKYBOX then continue end
		table.insert(motab,mo)
		table.insert(postab,{mo.x,mo.y,mo.z})
	end
	motab = shufflearray(motab)
	return motab,postab
end

//Get objects for position randomizing
local function GetObjectsRand()
	local motab = {}
	local postab = {}
	for mo in mobjs.iterate() do
		if mo.type == MT_PLAYER then continue end
		if mo.flags & flagblacklist != 0 then continue end
		if mo.type >= MT_TELEPORTMAN and mo.type <= MT_SKYBOX then continue end
		local isdone = false
		local newtab = {}
		while not isdone do
			newtab[1] = P_RandomRange(-32768,32767)*FRACUNIT
			newtab[2] = P_RandomRange(-32768,32767)*FRACUNIT
			if R_PointInSubsectorOrNil(newtab[1],newtab[2]) and P_CheckPosition(mo,newtab[1],newtab[2]) then isdone = true end
		end
		table.insert(motab,mo)
		table.insert(postab,newtab)
	end
	return motab,postab
end

//Get objects for object type swapping
local function GetObjectsForTypeSwap()
	local motab = {}
	for mo in mobjs.iterate() do
		if mo.type == MT_PLAYER or mo.type == MT_STARPOST or mo.type == MT_STEAM or mo.type == MT_SALOONDOOR or mo.type == MT_ROCKSPAWNER then continue end
		if mo.flags & flagblacklist != 0 then continue end
		if mo.type >= MT_TELEPORTMAN and mo.type <= MT_SKYBOX then continue end
		if mo.type >= MT_FLICKY_01 and mo.type <= MT_SECRETFLICKY_02_CENTER then continue end
		if vars.randomtype == 1 and mo.flags & limitedflags == 0 then continue end
		table.insert(motab,mo)
	end
	return motab
end

local function typeinsert(isProper,info,realtab,numtab)
	if isProper == true then
		table.insert(realtab,{#info,true})
		table.insert(numtab,{#info,true})
	elseif isProper == false then
		table.insert(realtab,{#info,false})
		table.insert(numtab,{#info,false})
	elseif isProper == "AAA" then
		table.insert(realtab,{#info,3})
		table.insert(numtab,{#info,false})
	end
end

local function SetObjectTypesForTypeSwap()
	local numtab = {}
	local realtab = {}
	local rtabgood = 0
	for f = 1,#mobjinfo-1 do
		local info = mobjinfo[f]
		if info.doomednum < 1 or #info == MT_PLAYER or #info == MT_STARPOST or #info == MT_STEAM or #info == MT_RING or #info == MT_SALOONDOOR or #info == MT_ROCKSPAWNER then typeinsert(false,info,realtab,numtab); continue end
		if info.flags & flagblacklist != 0 then typeinsert(false,info,realtab,numtab); continue end
		if #info >= MT_TELEPORTMAN and #info <= MT_SKYBOX then typeinsert(false,info,realtab,numtab); continue end
		if #info >= MT_FLICKY_01 and #info <= MT_SECRETFLICKY_02_CENTER then typeinsert(false,info,realtab,numtab); continue end
		if info.flags & limitedflags == 0 and #info != MT_RING then typeinsert("AAA",info,realtab,numtab); continue end
		rtabgood = $1 + 1
		typeinsert(true,info,realtab,numtab)
	end
	numtab = shufflearray(numtab)
	if #numtab == 0 or #realtab == 0 or rtabgood < 2 then return end
	if #numtab > 4096 or #realtab > 4096 then print("Table overflow"); return end
	for f = 1,#numtab do
		local tries = 0
		while numtab[f][2] == 3 or not numtab[f][2] do
			if tries > 256 then print("Type swap try limit breached"); break end
			numtab[f] = realtab[P_RandomRange(1,#realtab)]
			tries = tries + 1
		end
	end
	local xmobjs = {}
	for mo in mobjs.iterate() do
		if not realtab[mo.type][2] then continue end
		table.insert(xmobjs,mo)
	end
	for f = 1,#xmobjs do
		P_SpawnMobj(xmobjs[f].x,xmobjs[f].y,xmobjs[f].z,numtab[xmobjs[f].type][1])
		xmobjs[f].state = S_NULL
	end
end

//Get objects for object size randomizing
local function GetObjectsForSizeRand()
	local motab = {}
	for mo in mobjs.iterate() do
		if mo.type == MT_PLAYER or mo.type == MT_STARPOST or mo.type == MT_STEAM then continue end
		if mo.flags & flagblacklist != 0 then continue end
		if mo.type >= MT_TELEPORTMAN and mo.type <= MT_SKYBOX then continue end
		table.insert(motab,mo)
	end
	return motab
end

//Get random object type with a few restrictions
local function RandomObject()
	local good=false
	local minfo
	while not good do
		minfo = mobjinfo[P_RandomRange(MT_BLUECRAWLA,#mobjinfo-1)]
		if minfo.doomednum < 1 or #minfo == MT_STARPOST or #minfo == MT_SALOONDOOR or #minfo == MT_ROCKSPAWNER then continue end
		if vars.randomtype == 1 and minfo.flags & limitedflags == 0 then continue end
		if (#minfo >= MT_AXIS and #minfo <= MT_AXISTRANSFERLINE) or #minfo == MT_IDEYAANCHOR then continue end
		if #minfo >= MT_TELEPORTMAN and #minfo <= MT_SKYBOX then continue end
		good = true
	end
	return #minfo
end

local pickuptypes = {
	MT_RING,
	MT_COIN,
	MT_BLUESPHERE,
	MT_BOMBSPHERE,
	MT_REDTEAMRING,
	MT_BLUETEAMRING,
	MT_TOKEN,
	MT_REDFLAG,
	MT_BLUEFLAG,
	MT_EMBLEM,
	MT_EMERALD1,
	MT_EMERALD2,
	MT_EMERALD3,
	MT_EMERALD4,
	MT_EMERALD5,
	MT_EMERALD6,
	MT_EMERALD7,
}

local enemytypes = {}

local scenerytypes = {}

//Get random object type with a few restrictions (smart mode)
local function RandomObjectSmart(mo)
	local good=false
	local m=nil
	while not good do
		if isMoPickup(mo) or (mo.flags & MF_MONITOR) then 
			m = pickuptypes[P_RandomRange(1,#pickuptypes)]
		elseif (mo.flags & (MF_ENEMY|MF_BOSS)) then
			m = enemytypes[P_RandomRange(1,#enemytypes)]
		elseif (mo.flags & MF_SCENERY) then
			m = scenerytypes[P_RandomRange(1,#scenerytypes)]
		end
		good = true
	end
	return m
end

addHook("MapLoad",function(mapnum)
	if not vars.isenabled then return end
	if modeattacking or maptol & TOL_NIGHTS or maptol & TOL_XMAS then return end
	ObjLoading = true
	if vars.setposition == 1 then
		local motab,postab = GetObjectsSwap()
		for f = 1,#motab do
			local nx = postab[f][1]
			local ny = postab[f][2]
			local nz = postab[f][3]
			P_TeleportMove(motab[f],nx,ny,nz)
		end
	elseif vars.setposition == 2 then
		local motab,postab = GetObjectsRand()
		for f = 1,#motab do
			local nx = postab[f][1]
			local ny = postab[f][2]
			P_TeleportMove(motab[f],nx,ny,motab[f].z)
			motab[f].z = motab[f].floorz
			if motab[f].type == MT_RING then motab[f].z = $1 + 24*FRACUNIT end
		end
	end
	if vars.randomtype > 0 then
		if vars.randomtype == 2 then
			SetObjectTypesForTypeSwap()
		elseif vars.randomtype == 3 then
			local motab = GetObjectsForTypeSwap()
			local motabcount = #motab
			if #pickuptypes < 18 then
				for i = 1,#mobjinfo-1 do
					if mobjinfo[i].flags & MF_MONITOR then table.insert(pickuptypes,i) end
				end
			end
			if #enemytypes == 0 then
				for i = 1,#mobjinfo-1 do
					if mobjinfo[i].flags & (MF_ENEMY|MF_BOSS) then table.insert(enemytypes,i) end
				end
			end
			if #scenerytypes == 0 then
				for i = 1,#mobjinfo-1 do
					if mobjinfo[i].flags & MF_SCENERY and i != MT_SALOONDOOR then table.insert(scenerytypes,i) end
				end
			end
			for f = 1,motabcount do
				local objtype = RandomObjectSmart(motab[f])
				local newmo
				if objtype then
					newmo = P_SpawnMobj(motab[f].x,motab[f].y,motab[f].z,objtype) 
					motab[f].state = S_NULL
				end
			end
		else
			local motab = GetObjectsForTypeSwap()
			local motabcount = #motab
			for f = 1,motabcount do
				local newmo = P_SpawnMobj(motab[f].x,motab[f].y,motab[f].z,RandomObject())
				motab[f].state = S_NULL
			end
		end
	end
	if vars.randomsize > 0 then
		local motab = GetObjectsForSizeRand()
		local rmin = FRACUNIT
		local rmax = FRACUNIT
		for f = 1,#motab do
			if vars.randomsize == 1 then 
				rmin = FRACUNIT/2
				rmax = 2*FRACUNIT
			else
				rmin = FRACUNIT/8
				rmax = 5*FRACUNIT
			end
			if P_RandomChance(FRACUNIT/2) then
				motab[f].scale = P_RandomRange(rmin>>4,FRACUNIT>>4)<<4 //P_RandomRange has range limit of 65536
			else
				motab[f].scale = P_RandomRange(FRACUNIT>>4,rmax>>4)<<4
			end
		end
	end
	ObjLoading = false 
end)

addHook("MobjThinker",function(mo)
	if ObjLoading and leveltime < 2*TICRATE then return true end //Prevent P_MapStart: tmthing set! error
	if mo.urornddielag then mo.state = S_NULL; return end
	if not mo.otyperandom then return end
	mo.otyperandom = false
	mo.state = S_SPAWNSTATE
	mo.flags = mo.info.flags
end, MT_NULL)

//Remove lag on "Uniform" randomization
addHook("MobjSpawn",function(mo)
	if vars.randomtype == 2 then mo.urornddielag = true end
end, MT_BOXSPARKLE)

//Remove sprite errors on "Smart" randomization
addHook("MobjSpawn",function(mo)
	if vars.randomtype == 3 and mo.sprite >= SPR_ROIL and mo.sprite <= SPR_ROIP then mo.state = mobjinfo[MT_ROCKCRUMBLE1].spawnstate end
end)

//If object type randomizing is enabled, give players 2 seconds of invincibility after respawn
addHook("PlayerSpawn", function(p)
if netgame == false and vars.isenabled and vars.randomtype > 0 then
	p.powers[pw_flashing] = 2*TICRATE
end
end)